#include "python_if.h"

PyMethodDef RPNIterExpr_methods[] = {
/*
	{"eval", (PyCFunction)rpnexpr_eval, METH_FASTCALL, "Evaluate an expression"},
	{"reset_stack", (PyCFunction)rpnexpr_reset_stack, METH_NOARGS,
		"Reset stack memory storage (set all items to 0)"},
*/
	{"__getstate__", (PyCFunction)rpnif_getstate, METH_NOARGS,
		"Pickling method. Return a bytes repr of tokenized expression \
and the stack state."},
	{"__setstate__", (PyCFunction)rpnif_setstate, METH_O,
		"Unpickling method"},
	{NULL} //Sentinel
};

PyMemberDef RPNIterExpr_members[] = {
	{NULL}
};

PyGetSetDef RPNIterExpr_getset[] = {
	{"POS_LINEAR", rpnif_get_const, NULL,
	 "", NULL},
	{"POS_XY", rpnif_get_const, NULL,
	 "", NULL},
	{"POS_XDIM", rpnif_get_const, NULL,
	 "", NULL},

	{NULL}
};

PyTypeObject RPNIterExprType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	"pyrpn.RPNIterExpr",                     /* tp_name */
	sizeof(PyRPNIterExpr_t),                        /* tp_basicsize */
	0,                                               /* tp_itemsize */
	(destructor)rpnif_del, /* tp_dealloc */
	0,                                               /* tp_print */
	0,                                               /* tp_getattr */
	0,                                               /* tp_setattr */
	0,                                               /* tp_reserved */
	rpnif_repr,                                    /* tp_repr */
	0,                                               /* tp_as_number */
	0,                                               /* tp_as_sequence */
	0,                              /* tp_as_mapping */
	0,                                               /* tp_hash  */
	0,                                               /* tp_call */
	rpnif_str,                                               /* tp_str */
	0,                                               /* tp_getattro */
	0,                                               /* tp_setattro */
	0,                                               /* tp_as_buffer */
	Py_TPFLAGS_DEFAULT |
	Py_TPFLAGS_BASETYPE,   /* tp_flags */
	"RPN expression evaluator",                /* tp_doc */
	0,                                               /* tp_traverse */
	0,                                               /* tp_clear */
	0,                                               /* tp_richcompare */
	0,                                               /* tp_weaklistoffset */
	0,                                               /* tp_iter */
	0,                                               /* tp_iternext */
	RPNIterExpr_methods,                        /* tp_methods */
	RPNIterExpr_members,                        /* tp_members */
	RPNIterExpr_getset,                              /* tp_getset */
	0,                                               /* tp_base */
	0,                                               /* tp_dict */
	0,                                               /* tp_descr_get */
	0,                                               /* tp_descr_set */
	0,                                               /* tp_dictoffset */
	rpnif_init,          /* tp_init */
	0,                                               /* tp_alloc */
	rpnif_new,                            /* tp_new */
};

PyObject* rpnif_new(PyTypeObject *subtype, PyObject *args, PyObject* kwds)
{
	PyObject *ret, *err;
	PyRPNIterExpr_t *expr;
	ret = PyType_GenericNew(subtype, args, kwds);
	if((err = PyErr_Occurred()))
	{
		Py_DECREF(err);
		return ret;
	}
	expr = (PyRPNIterExpr_t*)ret;
	expr->rif = NULL;

	return ret;
}


int rpnif_init(PyObject *self, PyObject *args, PyObject *kwds)
{
	PyRPNExpr_t *expr_self;
	char *names[] = {"expression", "args_count", "stack_size", NULL};
	char err_str[256];
	const char *expr;
	long long int args_count, stack_size;

	expr_self = (PyRPNExpr_t*)self;

	stack_size = 16;
	expr_self->rpn = NULL;

	if(!PyArg_ParseTupleAndKeywords(args, kwds, "sL|L:RPNExpr.__init__", names, &expr,
		&args_count, &stack_size))
	{
		return -1;
	}

	if(strlen(expr) == 0)
	{
		PyErr_SetString(PyExc_ValueError,
			"RpnExpr.__init__() expect expression argument to be not empty");
		return -1;
	}

	if(args_count < 0 || args_count > 255)
	{
		snprintf(err_str, 128,
			"Argument count should be in [4..255] but %lld given",
			args_count);
		PyErr_SetString(PyExc_ValueError, err_str);
		return -1;
	}

	if(stack_size < 4 || stack_size > 255)
	{
		snprintf(err_str, 128,
			"Stack size should be in [0..255] but %lld given",
			stack_size);
		PyErr_SetString(PyExc_ValueError, err_str);
		return -1;
	}

	expr_self->rpn = malloc(sizeof(rpn_expr_t));
	if(!expr_self->rpn)
	{
		snprintf(err_str, 256,
			"Expression memory allocation error : %s",
			strerror(errno));
	}
	bzero(expr_self->rpn, sizeof(rpn_expr_t));

	if(rpn_expr_init(expr_self->rpn, stack_size, args_count) < 0)
	{
		snprintf(err_str, 256,
			"Expression init error : %s",
			expr_self->rpn->err_reason);
		PyErr_SetString(PyExc_ValueError, err_str);
		return -1;
	}

	expr_self->args = malloc(sizeof(rpn_value_t) * args_count);
	if(!expr_self->args)
	{
		snprintf(err_str, 256,
			"Error allocating arguments memory : %s",
			strerror(errno));
		return -1;
	}

	if(rpn_expr_compile(expr_self->rpn, expr))
	{
		PyErr_SetString(PyExc_ValueError, expr_self->rpn->err_reason);
		return -1;
	}

	return 0;
}

void rpnif_del(PyObject *self)
{
	PyRPNIterExpr_t *expr_self;

	expr_self = (PyRPNIterExpr_t*)self;
	if(expr_self->rif)
	{
		rpn_if_free(expr_self->rif);
		expr_self->rif = NULL;
	}
}

PyObject* rpnif_str(PyObject *self)
{
	Py_RETURN_NONE;
}

PyObject* rpnif_repr(PyObject *self)
{
	Py_RETURN_NONE;
}

PyObject* rpnif_getstate(PyObject *self, PyObject *noargs)
{
	Py_RETURN_NONE;
}

PyObject* rpnif_setstate(PyObject *self, PyObject *state_bytes)
{
	Py_RETURN_NONE;
}

PyObject* rpnif_get_const(PyObject *self, void* val)
{
	return Py_BuildValue("i", (int)(long int)42);
}
